/***************************************************************************
                          playlist.cpp  -  description
                             -------------------
    begin                : Die Apr 22 2003
    copyright            : (C) 2003 by J. Kofler
    email                : kaffeine@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/



#include <krandomsequence.h>
#include <kapplication.h>
#include <kglobal.h>
#include <kglobalsettings.h>
#include <kiconloader.h>
#include <kurldrag.h>
#include <klocale.h>
#include <kurlrequesterdlg.h>
#include <kfiledialog.h>
#include <kdirselectdialog.h>
#include <kmimetype.h>
#include <kstandarddirs.h>
#include <kprogress.h>
#include <kio/netaccess.h>
#include <kprinter.h>
#include <kmessagebox.h>
#include <kdebug.h>

#include <qdir.h>
#include <qlabel.h>
#include <qclipboard.h>
#include <qwhatsthis.h>
#include <qdragobject.h>
#include <qstringlist.h>
#include <qdom.h>
#include <qxml.h>
#include <qpainter.h>
#include <qpaintdevicemetrics.h>


#include <stdlib.h>
#include <pwd.h>

#include "playlist.h"


static QString GetTimeString(int msec);

PlayList::PlayList(const QString& mimes, QWidget *parent, const char *name )
 : QWidget(parent, name)
{

  setAcceptDrops(true);

  boxLayout = new QVBoxLayout(this);

  menuBar = new KMenuBar(this);
  menuBar->setSizePolicy ( QSizePolicy (QSizePolicy::MinimumExpanding, QSizePolicy::Fixed) );
  toolBar = new KToolBar(this);
  toolBar->setSizePolicy ( QSizePolicy (QSizePolicy::MinimumExpanding, QSizePolicy::Fixed) );
  
  
  list = new UrlListView(this);
  list->setAlternateBackground(QColor(0xFFD18B));
  list->setItemMargin(1);
  list->setMargin(2);
  list->setSelectionMode(QListView::Extended);
  list->addColumn("");
  list->addColumn(i18n("Title"));
  list->addColumn(i18n("Length"));
  list->addColumn("");
  list->addColumn("");
  list->setShowToolTips(true);
  list->setColumnWidthMode(0, QListView::Manual);
  list->setColumnWidthMode(1, QListView::Manual);
  list->setColumnWidthMode(2, QListView::Manual);
  list->setColumnWidthMode(3, QListView::Manual);
  list->setColumnWidthMode(4, QListView::Manual);
  
  list->setResizeMode(QListView::NoColumn);

  list->setDragEnabled(true);
  list->setAcceptDrops(true);
  list->setDropVisualizer(true);
  list->setItemsMovable(true);
  list->setSorting(-1);
  list->setItemsRenameable(true);
  list->setRenameable(1);
  list->setAllColumnsShowFocus(true);

  QWhatsThis::add(list, i18n("Playlist: Add entrys via menu or use drag and drop. Drag and drop is supported from Konqueror, Mozilla, ..."));

  boxLayout->addWidget(menuBar);
  boxLayout->addWidget(toolBar);
  boxLayout->addWidget(list);

  InitActions();
  InitMenus();
  
  connect(list, SIGNAL(dropped(QDropEvent*, QListViewItem*)), this, SLOT(slotDropEvent(QDropEvent*, QListViewItem*)));
  connect(list, SIGNAL(moved(QPtrList<QListViewItem>&, QPtrList<QListViewItem>&, QPtrList<QListViewItem>&)),
          this, SLOT(slotMove(QPtrList<QListViewItem>&, QPtrList<QListViewItem>&, QPtrList<QListViewItem>&)));
  connect(list, SIGNAL(doubleClicked(QListViewItem*)), this, SLOT(slotPlayDirect(QListViewItem*)));
  connect(list, SIGNAL(signalRemoveSelected()), this, SLOT(slotRemoveSelected()));
  connect(list, SIGNAL(signalPlayItem(QListViewItem*)), this, SLOT(slotPlayDirect(QListViewItem*)));
  
  mimeTypes = mimes;
  
  setCaption(i18n("Kaffeine Playlist"));

  isCurrentEntry = KGlobal::iconLoader()->loadIcon("player_play", KIcon::Small);
  infoPixmap = KGlobal::iconLoader()->loadIcon("info", KIcon::Small);
  cdPixmap = KGlobal::iconLoader()->loadIcon("mime_track", KIcon::Small);
  bookmarkPixmap = KGlobal::iconLoader()->loadIcon("bookmark", KIcon::Small);


  currentEntry = NULL;
  random = false; endless = false;

  dirLister = new KDirLister;
  connect(dirLister, SIGNAL(newItems(const KFileItemList&)),
          this,   SLOT(slotGetDirItems(const KFileItemList&)));
}


PlayList::~PlayList()
{  
 delete list;
}

//**** init actions and menus

void PlayList::InitActions()
{
  actionCollection = new KActionCollection(this);

  fileOpenPL   = new KAction(i18n("&Open Playlist"), "fileopen", Qt::CTRL|Qt::Key_O, this, SLOT(slotOpenPlaylist()), actionCollection, "openpl");
  fileSavePL   = new KAction(i18n("&Save Playlist"), "filesave", Qt::CTRL|Qt::Key_S, this, SLOT(slotSavePlaylist()), actionCollection, "savepl");

  fileExportHtml   = new KAction(i18n("&Html"),"mime_html", 0, this, SLOT(slotExportHtml()), actionCollection, "exporthtml");
  fileImportPls    = new KAction(i18n("&PLS"), "mime_txt", 0, this, SLOT(slotImportPls()), actionCollection, "importpls");
  fileImportNoatun = new KAction(i18n("&Noatun Playlist"),"noatun", 0, this, SLOT(slotImportNoatun()), actionCollection, "importnoatun");
  fileImportM3U    = new KAction(i18n("&M3U/XMMS"),"mime_txt", 0, this, SLOT(slotImportM3U()), actionCollection, "importm3u");
  fileImportAsx    = new KAction(i18n("&ASX"),"mime_html", 0, this, SLOT(slotImportAsx()), actionCollection, "importasx");
  fileClose        = new KAction(i18n("Close Playlist"), "fileclose", Qt::Key_P, this, SLOT(hide()), actionCollection, "closeplaylist");
  
  
  filePrint    = new KAction(i18n("&Print Playlist"), "fileprint",  0, this, SLOT(slotPrint()), actionCollection, "print");
  
  addFiles   = new KAction(i18n("&Add File(s)"), "mime_video", Qt::CTRL|Qt::Key_F, this, SLOT(slotAddFiles()), actionCollection, "addfiles");
  addUrl     = new KAction(i18n("Add &URL"), "package_network", Qt::CTRL|Qt::Key_U, this, SLOT(slotAddUrl()), actionCollection, "addurl");
  addDir     = new KAction(i18n("Add &Directory"), "folder_video", Qt::CTRL|Qt::Key_D, this, SLOT(slotAddDir()), actionCollection, "adddir");
  addPaste  = new KAction(i18n("&Insert Selection as URL"), "editpaste", Qt::Key_Insert, this, SLOT(slotPaste()), actionCollection, "paste");
  addPaste->setWhatsThis(i18n("Insert the selected text of another X-window (e.g. Konqueror or Mozilla window) as a URL. Will be checked if valid."));

  editClear  = new KAction(i18n("&Clear Playlist"), "editclear", Qt::CTRL|Qt::Key_C, this, SLOT(slotClearList()), actionCollection, "clearlist");
  editRemove = new KAction(i18n("&Remove Selected"), "editdelete", Qt::Key_Delete, this, SLOT(slotRemoveSelected()), actionCollection, "remove");

  sortByTitle =  new KAction(i18n("Sort by titles"), "gear", Qt::Key_S, this, SLOT(slotSortByTitle()), actionCollection, "sorttitleascending");
  sortByType  =   new KAction(i18n("Sort by types"), "gear", Qt::SHIFT | Qt::Key_S , this, SLOT(slotSortByType()), actionCollection, "sorttypeeascending");
}


void PlayList::InitMenus()
{

  fileMenu = new KPopupMenu(this);
  addMenu  = new KPopupMenu(this);
  editMenu = new KPopupMenu(this);
  importMenu = new KPopupMenu(this);
  exportMenu = new KPopupMenu(this);
  sortMenu = new KPopupMenu(this);

  fileExportHtml->plug(exportMenu);
 
  fileImportNoatun->plug(importMenu);
  fileImportM3U->plug(importMenu);
  fileImportPls->plug(importMenu);
  fileImportAsx->plug(importMenu);

  fileOpenPL->plug(fileMenu);
  fileOpenPL->plug(toolBar);
  fileSavePL->plug(fileMenu);
  fileSavePL->plug(toolBar);
 
  fileMenu->insertSeparator();
  fileMenu->insertItem( KGlobal::iconLoader()->loadIconSet("fileimport", KIcon::Small),
                        i18n("&Import..."), importMenu);
  fileMenu->insertItem( KGlobal::iconLoader()->loadIconSet("fileexport", KIcon::Small),
                        i18n("&Export to"), exportMenu);
  filePrint->plug(fileMenu);
  fileMenu->insertSeparator();

  fileClose->plug(fileMenu);
  fileClose->plug(toolBar);
  toolBar->insertLineSeparator();
  
  addDir->plug(addMenu);
  addDir->plug(toolBar);
  addFiles->plug(addMenu);
  addFiles->plug(toolBar);
  addUrl->plug(addMenu);
  addUrl->plug(toolBar);
  addPaste->plug(addMenu);
  addPaste->plug(toolBar);
  toolBar->insertLineSeparator();

  sortByTitle->plug(sortMenu);
  sortByTitle->plug(toolBar);
  sortByType->plug(sortMenu);

  editMenu->insertItem(i18n("&Sort Playlist"), sortMenu);
  editRemove->plug(editMenu);
  editRemove->plug(toolBar);
  editClear->plug(editMenu);
  editClear->plug(toolBar);
  
  menuBar->insertItem(i18n("&File"), fileMenu);
  menuBar->insertItem(i18n("&Add"), addMenu);
  menuBar->insertItem(i18n("&Edit"), editMenu);
  
  toolBar->insertSeparator();
  
  toolBar->insertLined("", 123, SIGNAL(returnPressed(const QString&)), this, SLOT(slotFindText(const QString&)),
                       true, i18n("Find Expression in Titles"));
  toolBar->alignItemRight(123);

  
}

/******************************************
   get the urls
*******************************************/

QListViewItem* PlayList::GetCurrent()
{
  if (random)
  {
    if (currentRandomListEntry == -1) return NULL;
    SetCurrentEntry(randomList.at(currentRandomListEntry));
    return currentEntry;
  }  

  if (!currentEntry)
    if (list->childCount()>0)
    {
      SetCurrentEntry(list->firstChild());
    }
    else
     return NULL;  

   return currentEntry;
}


QListViewItem* PlayList::GetNext()
{
  if (!currentEntry) return GetCurrent();

  if (random)
  {
  if ((currentRandomListEntry+1) < (int)randomList.count())
      currentRandomListEntry += 1;
     else
     {
       if (endless)
         currentRandomListEntry = 0;
        else
         return NULL;
     }
    SetCurrentEntry(randomList.at(currentRandomListEntry));
    return currentEntry;     
   }

  QListViewItem* tmpItem;
  tmpItem = currentEntry->itemBelow();
  
  if (tmpItem)
  {
    SetCurrentEntry(tmpItem);
    return currentEntry;
  }
  else
  {
    if (endless)
      {
        SetCurrentEntry(list->firstChild());
        return currentEntry;
      }  
     else
      return NULL;
  }  
}


QListViewItem* PlayList::GetPrevious() 
{
  if (!currentEntry) return GetCurrent();

  if (random)
  {
    if (currentRandomListEntry > 0)
      currentRandomListEntry -= 1;
     else
     {
       if (endless)
         currentRandomListEntry = randomList.count()-1;
        else
         return NULL;
     }
    SetCurrentEntry(randomList.at(currentRandomListEntry));
    return currentEntry;
   }

  QListViewItem* tmpItem;
  tmpItem = currentEntry->itemAbove();

  if (tmpItem)
  {
    SetCurrentEntry(tmpItem);
    return currentEntry;
  }
  else
  {
    if (endless)
     {
       SetCurrentEntry(list->lastItem()); 
       return currentEntry;
     }  
     else
      return NULL;
  }  
}


//********** set current entry (with play icon) ****************

void PlayList::SetCurrentEntry(QListViewItem* item)
{
  if (currentEntry)
    currentEntry->setPixmap(1, QPixmap());
  
  item->setPixmap(1, isCurrentEntry);
  currentEntry = item;
  currentChild = NULL;
  list->setCurrentItem(currentEntry);

  list->ensureVisible(10, list->itemPos(currentEntry), 10, 30);
  
}

/***************************************************
 *          create new items                       *
 ***************************************************/            

QListViewItem* PlayList::CreateItem(QListViewItem* after, QString url, QString mime, QString title,
   QListViewItem* parent, QString length, QString info)
{
  QListViewItem* tmp = NULL;
        
  if (list->findItem(url, 4))
    return NULL;

  if (title.isNull())
  {
    KURL kurl(url);
  
    if (kurl.isLocalFile())
       title = kurl.fileName();
     else
       title = kurl.prettyURL();
   }

   if (mime.isNull())
   {
     KMimeType::Ptr getMime = KMimeType::findByURL(url);
     mime = getMime->name();
   }
     
   if (parent)
   {
     tmp = new KListViewItem(parent, after, mime, title, length, info, url);
     if (!tmp) return NULL;
   }
   else
   {
     tmp = new KListViewItem(list, after, mime, title, length, info, url);
     if (!tmp) return NULL;
   }  
     
   if (mime == "autoplay-plugin")
      tmp->setPixmap(0, cdPixmap);
     else
      tmp->setPixmap(0, KMimeType::mimeType(mime)->pixmap(KIcon::Small));
      
   tmp->setPixmap(4, bookmarkPixmap);
   if (!info.isEmpty()) tmp->setPixmap(3, infoPixmap);
   return tmp;
  
}     

 

/*****************************************
  add urls to playlist
*****************************************/  

void PlayList::AddAutoplayUrls(const QStringList& urls, QListViewItem* after)
{
  QListViewItem* tmp = NULL;
  
  for (unsigned int i=0; i<urls.count(); i++)
  {
    tmp = CreateItem(after, urls[i], "autoplay-plugin");
    if (tmp) after = tmp;
  }

  if (random) CreateRandomList();
}


void PlayList::Add(const KURL& url, QListViewItem* after)
{
  Add(KURL::List(url), after);
}


void PlayList::Add(const KURL::List& urls, QListViewItem* after)
{
  QStringList skipped;
  QListViewItem* tmp = NULL;

  KProgressDialog* pDialog = new KProgressDialog(0, "checkprogress", QString::null, i18n("Checking URL(s)..."));
  pDialog->setAutoClose(true);
  pDialog->progressBar()->setTotalSteps(urls.count()-1);

  QString ext;
  QString fname;
  
  for (uint i=0; i<urls.count(); i++)
  {
     //********** check if autoplay plugin ************

     fname = urls[i].fileName();
     uint a = 0;
     while (a < autoplayList.count())
     {
       if (fname == autoplayList[a])
       {
         emit signalAutoplayPlugin(fname);
         delete pDialog;
         return;  
       }
       a+=2;
     }    

     //*********** check mime type ******************

     ext = urls[i].url().remove(0 ,urls[i].url().length()-3).lower();
     
     KMimeType::Ptr mime = KMimeType::findByURL(urls[i].path()); ///path() ????????????????
     int c = mimeTypes.contains(mime->name()) || QString("wma wmv WMA WMV").contains(ext); /* maybe mime type of *.wma, *.wmv are not detected correctly*/
     
     if (c == 0)
     {

     //  check if kaffeine/noatun/asx playlist
            
       if ((mime->name() == "text/plain") || (mime->name() == "text/xml") || (mime->name() == "text/html")
           || (mime->name() == "audio/x-scpls") || (mime->name() == "audio/x-mpegurl") || (ext == "asx") )
       {
         kdDebug(555) << "Check for kaffeine/noatun/m3u/pls/asx playlist\n";
         QString localFile;
         if (KIO::NetAccess::download(urls[i], localFile))
         {
           QFile file(localFile);
           file.open(IO_ReadOnly);

           QTextStream stream(&file);
           QString firstLine = stream.readLine();
           QString secondLine = stream.readLine();
           file.close();
          
           if (secondLine.contains("kaffeine", false))
           {
             kdDebug(555) << "Try loading kaffeine playlist\n";
             if (!LoadPlaylist(localFile, after))
                skipped << urls[i].prettyURL() + " (" + mime->name() + " )";
             continue;
           }
           if (secondLine.contains("noatun", false))
           {
             kdDebug(555) << "Try loading noatun playlist\n";
             if (!LoadNoatunPlaylist(localFile, after))
                skipped << urls[i].prettyURL() + " (" + mime->name() + " )";
             continue;
           }
           if (firstLine.contains("asx", false))
           {
             kdDebug(555) << "Try loading asx playlist\n";
             if (!LoadAsxPlaylist(localFile, after))
                skipped << urls[i].prettyURL() + " (" + mime->name() + " )";
             continue;
           }
           
           if (firstLine.contains("[playlist]", false))
           {
             kdDebug(555) << "Try loading pls playlist\n";
             if (!LoadPlsPlaylist(localFile, after))
                skipped << urls[i].prettyURL() + " (" + mime->name() + " )";
             continue;
           }
           QString url = urls[i].url();
           if (ext == "m3u")  //indentify by extension
           {
             kdDebug(555) << "Try loadng m3u playlist\n";
             if (!LoadM3UPlaylist(localFile, after))
                skipped << urls[i].prettyURL() + " (" + mime->name() + " )";
             continue;
           }
           
         } 
       }
       if (urls[i].isLocalFile())  // only skip local files
       {
         skipped << urls[i].prettyURL() + " (" + mime->name() + " )";
         continue;
       }
       else
         mime = KMimeType::mimeType("unknown");
     }

     tmp = CreateItem(after, urls[i].prettyURL(), mime->name());
     if (!tmp)
        skipped << urls[i].prettyURL() << i18n(" (still exist) ");
       else
        after = tmp;
        
   if (pDialog->wasCancelled()) break;
   pDialog->progressBar()->setProgress(i);
     
  }

  delete pDialog;
  if (random) CreateRandomList();

  if (skipped.count())
  //  KMessageBox::informationList(0, i18n("Following URL(s) are not accepted: "), skipped); // qt3.1 only
  KMessageBox::warningContinueCancelList(this, i18n("Following URL(s) are not accepted: "), skipped, QString::null, KStdGuiItem::ok());
}



//********** insert child urls (ram, etc) ***********

void PlayList::slotAddMrl(const QString& mrl)
{
  currentEntry->setOpen(true);

  QListViewItem* tmp = CreateItem(currentChild, mrl, QString::null, QString::null, currentEntry);
  if (tmp) currentChild = tmp;
  
  currentEntry->setText(3, QString(i18n(" Contents %1 URL(s)" )).arg(currentEntry->childCount()));
  if (random) CreateRandomList();    
}

//********************************************************


void PlayList::slotPlayDirect(QListViewItem* item)
{
  SetCurrentEntry(item);
  if (random)
    currentRandomListEntry = randomList.find(item);
  emit signalPlayDirect();
}  
  

void PlayList::SetEndless(bool e)
{
  endless = e;
}

void PlayList::SetRandomPlay(bool r)
{
  random = r;
  if (random) CreateRandomList();
}


void PlayList::CreateRandomList()
{
  randomList.clear();
  currentRandomListEntry = 0;

  QListViewItem* tmpItem;
  tmpItem = list->firstChild();


  
  while (tmpItem)
  {
    randomList.append(tmpItem);
    tmpItem = tmpItem->itemBelow();
  }

 if (!(randomList.count()>0))
 {
   currentRandomListEntry = -1;
   return;
 }  
 
 KRandomSequence r(KApplication::random());
 r.randomize(&randomList);
}


void PlayList::slotClearList()
{
  list->clear();
  randomList.clear();
  currentEntry = NULL;
  currentRandomListEntry = -1;
}

 
void PlayList::slotMove(QPtrList<QListViewItem>& items, QPtrList<QListViewItem>&, QPtrList<QListViewItem>& afterNow)
{
  for (unsigned int i=0; i<items.count(); i++)
   items.at(i)->moveItem(afterNow.at(i));
}


void PlayList::slotDropEvent(QDropEvent* dev, QListViewItem* after)
{
   KURL::List uris;
   
   if (KURLDrag::decode(dev, uris))
   {
     Add(uris, after);
   }
   else
   if (strcmp(dev->format(), "text/x-moz-url") == 0)    // for mozilla drops
   {
     QByteArray data = dev->encodedData("text/plain");
     QString md(data);

 
     Add(md, after);
       
   }
  
}

void PlayList::slotSetMetaInfoString(const QString& mi)
{
  metaInfoString = mi;
}  


void PlayList::slotAddFiles()
{
  KURL::List urlList;

  urlList = KFileDialog::getOpenURLs(QDir::homeDirPath(), QString::null, 0, i18n("Add File(s)"));


  if (!urlList.count()) return;

  Add(urlList, list->lastItem());
}


void PlayList::slotAddUrl()
{

 KURL kurl = KURLRequesterDlg::getURL(QString::null, 0, i18n("Add URL"));

 if (!kurl.isValid()) return;

 Add(kurl, list->lastItem());

}

void PlayList::slotAddDir()
{
  KURL path = KDirSelectDialog::selectDirectory(QDir::homeDirPath(), false, 0, i18n("Add Directory"));

  if (path.isValid())
    dirLister->openURL(path);
}

void PlayList::slotGetDirItems(const KFileItemList& items)
{
  KFileItemListIterator it(items);   

  KFileItem* cur;
  KURL::List kurlList;

  cur = it.current();

  while (cur)
  {
    kurlList.append(cur->url());
    ++it;
    cur = it.current();
  }  

  Add(kurlList, list->lastItem());
  
  dirLister->stop();
}

/************ sort playlist ******************/

void PlayList::slotSortByTitle()
{
  list->setSorting(1);
  list->sort();
  list->setSorting(-1);
}


void PlayList::slotSortByType()
{
  list->setSorting(0);
  list->sort();
  list->setSorting(-1);
}


/********************************************/

void PlayList::slotRemoveSelected()
{
  QPtrList<QListViewItem> selected;

  if (currentEntry)
    if (currentEntry->isSelected())
     {
        currentEntry = NULL;
        currentRandomListEntry = -1;
     }   

  selected = list->selectedItems();

  
  for(unsigned int i=0; i<selected.count(); i++)
  {
    kdDebug(555) << "Remove " << selected.at(i)->text(4) << "\n";
    delete selected.at(i);
  } 
 
  if (random) CreateRandomList();  
}

/**************************************************
             load xml playlist
******************************************************/         

class MyXMLParser : public QXmlDefaultHandler
{

public:
  PlayList *playList;
  QListViewItem* tmp;
  QListViewItem* after;
  bool isKaffeinePlaylist;
  

MyXMLParser(PlayList *pl): playList(pl), tmp(NULL), after(NULL), isKaffeinePlaylist(false)
{
}

bool startElement(const QString&, const QString&,
            const QString &qname, const QXmlAttributes &att)
 {
                                                 
  if (qname == "playlist")
  if (att.value("client") == "kaffeine")
   {
     isKaffeinePlaylist = true;
     return true;
   } 
   else
   {
     return false;  
   }

  
  if (qname != "entry") return true;
  
  tmp = playList->CreateItem(after, att.value("url"), att.value("mime"), att.value("title"), NULL,
                   att.value("length"), att.value("info"));

  if (tmp) after = tmp;
   
  return true;
 }

};

bool PlayList::LoadPlaylist(const QString& pl, QListViewItem* after)
{
  QFile file(pl);
  if (!file.open(IO_ReadOnly)) return false;
 
  QXmlInputSource source(&file);
  QXmlSimpleReader reader;

  MyXMLParser parser(this);
  parser.after = after;
  reader.setContentHandler(&parser);
  reader.parse(source);
  file.close();

  return parser.isKaffeinePlaylist; 
}

void PlayList::slotOpenPlaylist()
{
  
  QString path = KFileDialog::getOpenFileName(QDir::homeDirPath() + "/.kaffeine", QString::null, 0, i18n("Open Playlist"));

  if (path.isEmpty()) return;
  
  slotClearList();

  if (!LoadPlaylist(path, NULL))
  {
    KMessageBox::sorry(this, QString(i18n("No Kaffeine Playlist!")));
    lastPlaylist = QString();
    return;
  }

  if (random) CreateRandomList();

  lastPlaylist = path;
  list->SetCleared(false);
  
}

/****************************************************
               load noatun playlist
*******************************************************/           

class NoatunXMLParser : public QXmlDefaultHandler
{

public:
  PlayList *playList;
  QListViewItem* tmp;
  QListViewItem* after;
  bool isNoatunPlaylist;


  
NoatunXMLParser(PlayList *pl): playList(pl), tmp(NULL), after(NULL), isNoatunPlaylist(false)
{
}

bool startElement(const QString&, const QString &,
            const QString &qname, const QXmlAttributes &att)
 {
                                                 
   if (qname == "playlist")
   if (att.value("client") == "noatun")
   {
    isNoatunPlaylist = true;
    return true;
   }
   else
   {
     return false;
   }

  if (qname != "item") return true;

  QString timeString = QString::null;
  QString title = QString::null;

  if (!att.value("title").isNull())
  {
    title = playList->metaInfoString.arg(att.value("title")).arg(att.value("author")).arg(att.value("album"));
  }  
  
  bool ok;
  int time = 0;
  time = att.value("length").toInt(&ok);
  
  if ((ok) && (time > 0)) timeString = GetTimeString(time);
  
  tmp = playList->CreateItem(after, att.value("url"), QString::null, title, NULL, timeString);

  if (tmp) after = tmp;
  
  return true;
 }

};



bool PlayList::LoadNoatunPlaylist(const QString& pl, QListViewItem* after)
{
  QFile file(pl);
  if (!file.open(IO_ReadOnly)) return false;

  QXmlInputSource source(&file);
  QXmlSimpleReader reader;

  NoatunXMLParser parser(this);
  parser.after = after;
  reader.setContentHandler(&parser);
  reader.parse(source);
  file.close();

  return parser.isNoatunPlaylist;
}


void PlayList::slotImportNoatun()
{
  QString path = KFileDialog::getOpenFileName(QDir::homeDirPath(), QString::null, 0, i18n("Import Noatun Playlist"));

  if (path.isEmpty()) return;

  slotClearList();
  
  if (!LoadNoatunPlaylist(path, NULL))
  {
    KMessageBox::sorry(this, QString(i18n("No Noatun Playlist!")));
    return;
  }

  if (random) CreateRandomList();  

}

/*****************************************************
                load asx playlist
******************************************************/

bool PlayList::LoadAsxPlaylist(const QString& pl, QListViewItem* after)
{
  QFile file(pl);
  if (!file.open(IO_ReadOnly)) return false;

  QDomDocument doc;
  doc.setContent(&file);
  QDomElement root = doc.documentElement();

  QString url;
  QString title;

  QListViewItem* tmp = NULL;

  if (root.nodeName().lower() != "asx") return false;

  QDomNode node = root.firstChild();
  QDomNode subNode;
  QDomElement element;
  KURL::List kurlList;

  while (!node.isNull())
  {
    url = QString::null;
    title = QString::null;
    if (node.nodeName().lower() == "entry")
    {
      subNode = node.firstChild();
      while (!subNode.isNull())
      {
        if ((subNode.nodeName().lower() == "ref") && (subNode.isElement()))
        {
          element = subNode.toElement();
          if (element.hasAttribute("href"))
             url = element.attribute("href");
          else if (element.hasAttribute("HREF"))
             url = element.attribute("HREF"); 
         }

        if ((subNode.nodeName().lower() == "title") && (subNode.isElement()))
         {
           title = subNode.toElement().text();
         }
        
        subNode = subNode.nextSibling();
      }

      if (!url.isNull())
      {
        tmp = CreateItem(after, url, QString::null, title);
        if (tmp) after = tmp;
      }  
      
    }
    node = node.nextSibling();
  }

  file.close();
  
  return true;
}


void PlayList::slotImportAsx()
{
  QString path = KFileDialog::getOpenFileName(QDir::homeDirPath(),
                 i18n("*.asx *.ASX|ASX-Files\n*.*|All Files"), 0, i18n("Import ASX Playlist"));

  if (path.isEmpty()) return;

  slotClearList();

  if (!LoadAsxPlaylist(path, NULL))
  {
    KMessageBox::sorry(this, QString(i18n("No ASX File: %1")).arg(path));
    return;
  }

  if (random) CreateRandomList();  

}


/******************************************************
                    load m3u playlist
******************************************************/       

bool PlayList::LoadM3UPlaylist(const QString& pl, QListViewItem* after)
{
  QFile file(pl);
  if (!file.open(IO_ReadOnly)) return false;


  QTextStream stream(&file);

//  if (stream.readLine().upper() != "#EXTM3U") return false;
  
  QString url;
  int time;
  bool ok;
  QString timeString;
  QString title;
  KURL kUrl;
  QListViewItem* tmp = NULL;

  bool foundAnyUrl = false;

  
  while (!stream.eof())
  {
    url = stream.readLine();
  
    if (url.left(1) == "#")
    {
      if (url.left(7).upper() == "#EXTINF")
        {
          time = 0;
          title = QString::null;
          timeString = QString::null;

          url = url.remove(0,8);

          time = url.section(",", 0,0).toInt(&ok);
          if ((ok) && (time > 0)) timeString = GetTimeString(time*1000); //msec

          title = url.section(",",1,1);

          url = stream.readLine();

          tmp = CreateItem(after, url, QString::null, title, NULL, timeString);
          if (tmp) after = tmp;

          foundAnyUrl = true;

         }
         continue;
      }
       

    kUrl = KURL(url);
    if (kUrl.isValid())
    {
      foundAnyUrl = true;

      tmp = CreateItem(after, url);
      if (tmp) after = tmp;
     }
  }    

  file.close();

  if (foundAnyUrl)
    return true;
   else 
    return false;
}  


void PlayList::slotImportM3U()
{
  QString path = KFileDialog::getOpenFileName(QDir::homeDirPath(),
                 i18n("*.m3u *.M3U|M3U-Files\n*.*|All Files"), 0, i18n("Import M3U/XMMS Playlist"));

  if (path.isEmpty()) return;               

  slotClearList();

  if (!LoadM3UPlaylist(path, NULL))
  {
    KMessageBox::sorry(this, QString(i18n("No M3U File: %1")).arg(path));
    return;
  }

  if (random) CreateRandomList();    
  
}

/***************************************************
                load pls playlist
**************************************************/    

bool PlayList::LoadPlsPlaylist(const QString& pl, QListViewItem* after)
{
  QFile file(pl);
  if (!file.open(IO_ReadOnly)) return false;

  QListViewItem* tmp = NULL;
  
  QTextStream stream(&file);

  if (stream.readLine().upper() != "[PLAYLIST]") return false;
  
  bool ok;
  QString numEntrys = stream.readLine().remove(0,16);
  int entrys = 0;
  entrys = numEntrys.toInt(&ok);
  if (!(ok) || !(entrys>0)) return false;

  QString* titles = new QString[entrys];
  QString* files = new QString[entrys];
  QString* length = new QString[entrys];

  for (int i=0; i<entrys; i++)
  {
    titles[i] = QString::null;
    files[i] = QString::null;
    length[i] = QString::null;
  }  
  
  QString line;
  int num;
  int time;
  

  while (!stream.eof())
  {
    line = stream.readLine();
   
    if (line.left(4).lower() == "file")
    {
      line = line.remove(0,4);

      num = line.left(line.find("=")).toInt(&ok);
      if ((!ok) && (num>entrys)) continue;

      files[num-1] = line.remove(0, line.find("=") + 1);

      continue;
    }

    if (line.left(5).lower() == "title")
    {
      line = line.remove(0,5);

      num = line.left(line.find("=")).toInt(&ok);
      if ((!ok) && (num>entrys)) continue;

      titles[num-1] = line.remove(0, line.find("=") + 1);

      continue;
    }

    if (line.left(6).lower() == "length")
    {
      line = line.remove(0,6);

      num = line.left(line.find("=")).toInt(&ok);
      if ((!ok) && (num>entrys)) continue;

      time = 0;
      time = line.remove(0, line.find("=") + 1).toInt(&ok);
      if ( !(ok) || !(time > 0) ) continue;

      length[num-1] = GetTimeString(time*1000);

    }  
    
  }

  file.close();

  for (int i=0; i<entrys; i++)
  {
    if (files[i].isNull()) continue;
    tmp = CreateItem(after, files[i], QString::null, titles[i], NULL, length[0]);
    if (tmp) after = tmp;
  }  

  delete [] titles;
  delete [] files;
  delete [] length;

  return true;
}  

void PlayList::slotImportPls()
{
  QString path = KFileDialog::getOpenFileName(QDir::homeDirPath(),
                  i18n("*.pls *. PLS|PLS-Files\n*.*|All Files"), 0, i18n("Import PLS Playlist"));

  if (path.isEmpty()) return;
                  
  slotClearList();
  
  if (!LoadPlsPlaylist(path, NULL))
    KMessageBox::sorry(this, QString(i18n("No PLS File: %1")).arg(path));
  
}


//*********** save xml playlist *****************

bool PlayList::SavePlaylist(const QString& pl)
{
  QDomDocument doc("playlist");
  doc.setContent(QString("<!DOCTYPE XMLPlaylist>"));
  QDomElement root = doc.createElement("playlist");
  root.setAttribute("client", "kaffeine");
  doc.appendChild(root);

  QDomElement entry;
  QListViewItem* tmp = list->firstChild();

  while (tmp)
  {
    if (!tmp->parent())   //dont save sub-urls
    {
      entry = doc.createElement("entry");

      entry.setAttribute("title", tmp->text(1));
      entry.setAttribute("url", tmp->text(4));
      entry.setAttribute("mime", tmp->text(0));
      entry.setAttribute("length", tmp->text(2));
      entry.setAttribute("info", tmp->text(3));

      root.appendChild(entry);
    }

    tmp = tmp->itemBelow();
  }


  QFile file(pl);
  if (!file.open(IO_WriteOnly)) return false;
  QTextStream stream(&file);
  stream.setEncoding(QTextStream::UnicodeUTF8);

  stream << doc.toString();

  file.close();
  
  return true;
}


void PlayList::slotSavePlaylist()
{
  QString startPath = QDir::homeDirPath() + "/.kaffeine";
  if (!lastPlaylist.isEmpty()) startPath = lastPlaylist;
  
  QString path = KFileDialog::getSaveFileName(startPath, QString::null, 0, i18n("Save Playlist"));

  if (path.isEmpty()) return;
  
  if (SavePlaylist(path))
  {
    lastPlaylist = path;
    list->SetCleared(false);
  }  
  else
    {
      kdDebug(555) << "Can't save Playlist\n";
      lastPlaylist = QString();
    }  
}  


//**********************************************
  
void PlayList::slotGetLengthInfo(const QString& time)
{
  currentEntry->setText(2,time);
}

void PlayList::slotGetMetaInfo(const QString& metaInfo)
{
  if (!currentEntry->pixmap(3))
    currentEntry->setText(1, metaInfo);
}



void PlayList::slotGetStreamInfo(const QString& streamInfo)
{
  currentEntry->setText(3, streamInfo);
  if (!currentEntry->pixmap(3))
  {    
    currentEntry->setPixmap(3, infoPixmap);
  }
}


void PlayList::slotFindText(const QString& text)
{
  if (text.isEmpty()) return;
  
  QListViewItem* tmp = list->firstChild();
  int answer;
  
  while (tmp)
  {
    /* search in titles und urls, case unsensitive */
    if ( tmp->text(1).contains(text, false) || tmp->text(4).contains(text, false) )
    {
      list->clearSelection();
      list->ensureVisible(10, list->itemPos(tmp), 10, 30);
      tmp->setSelected(true);
      list->triggerUpdate();
      
      answer = KMessageBox::questionYesNo(this, i18n("Find next occurence of ") + "'" + text + "' ?");
      if (answer == KMessageBox::No) break;
    }  
  
    tmp = tmp->itemBelow();
  }
  setFocus();   /*take the focus from the line editor*/
}  



void PlayList::SetLastPlaylist(const QString& pl)
{
  lastPlaylist = pl;
}
  
const QString& PlayList::GetLastPlaylist() const
{
  return lastPlaylist;
}
  
void PlayList::ReloadLastPlaylist()
{
  if (lastPlaylist.isEmpty()) return;

  LoadPlaylist(lastPlaylist, NULL);

  if (random) CreateRandomList();

  list->SetCleared(false);
}


void PlayList::SaveCurrentPlaylist()
{
  if ( (lastPlaylist.isEmpty()) || (list->GetCleared()) ) return;

  kdDebug(555) << "Save current Playlist\n";
  SavePlaylist(lastPlaylist);
}
  

void PlayList::slotPaste()
{
  QClipboard *cb = QApplication::clipboard();
  
//  QString text = cb->text(QClipboard::Selection);  // qt3.1 only

  cb->setSelectionMode(true);
  QString text = cb->text();

  KURL kurl;
  kurl.setPath(text);

  if (text)
    Add(kurl, list->lastItem());
  
}


/***********************************************************
                  export to html
************************************************************/                  

void PlayList::slotExportHtml()
{
  bool bgcol = false;

  QString path = KFileDialog::getSaveFileName(QDir::homeDirPath(),
                                 i18n("*.html *.htm|Html-Files\n*.*|All Files"),
                                 0, i18n("Export Playlist to Html"));
  QFile file(path);
  if (!file.open(IO_WriteOnly)) return;

  if (!QFile::exists(QDir::homeDirPath() + "/.kaffeine/kmp-playlist.png") )
  {
    QStringList dirs = KGlobal::dirs()->findDirs("data","kaffeine");
    QString logoPath = dirs[0];  // global kde data path
    logoPath.append("kmp-playlist.png");
    QImage logo;
    logo.load(logoPath);
    logo.save(QDir::homeDirPath() + "/.kaffeine/kmp-playlist.png", "PNG");
  }

  //******get real name of user and date
  QString realName;
  QString date;

  QString uname = getenv("USER");
  if (!uname)
    uname = getenv("LOGNAME");
  if (uname)
  {
    struct passwd* pw = getpwnam(uname);
    realName = pw->pw_gecos;
  }
    
  date = KGlobal::locale()->formatDate(QDate::currentDate(), true);
  //**********
  

  QTextStream stream(&file);
  stream.setEncoding(QTextStream::UnicodeUTF8);

  stream << "<!DOCTYPE public \"-//w3c//dtd html 4.01 transitional//en\"\n";
  stream << "\"http://www.w3.org/TR/html4/loose.dtd\">\n";
  stream << "<html>\n";
  stream << "<head>\n";
  stream << "<title>Kaffeine Playlist</title>\n";
  stream << "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF8\">\n";
  stream << "<meta name=\"GENERATOR\" content=\"Kaffeine Media Player\">\n";
  stream << "</head>\n";
  stream << "<body bgcolor=\"white\" text=\"black\">\n";
  stream << "<p align=\"center\">\n";
  stream << "<img src=\"" << QDir::homeDirPath() + "/.kaffeine/kmp-playlist.png" << "\" width=\"467\" height=\"168\" hspace=\"10\" border=\"0\">\n";
  stream << "</p><p align=\"center\">" << realName << ", " << date << "</p>\n";
  stream << "<table border=\"0\" align=\"center\" cellpadding=\"2\">\n";
  stream << "<tbody>\n";


  QListViewItem* tmp = list->firstChild();

  while (tmp)
  {
    if (!tmp->parent())   //dont save sub-urls
    {
      if (bgcol)
      {
        stream << "<tr bgcolor=\"#FFD18B\">\n";
        bgcol = false;
      }
      else
      {
        stream << "<tr>\n";
        bgcol = true;
      }    
      stream << "<td>" << i18n("Title") << ":<br>" << i18n("URL") << ":<br>" + i18n("Length") << ":<br>" << i18n("Stream") << ":</td>\n";

      stream << "<td>\n";
      stream << "<b>" << tmp->text(1) << "</b><br>\n"; //title
      stream << "<a href =\"" << tmp->text(4) << "\">" << tmp->text(4) << "</a><br>\n"; //url
      stream << tmp->text(2) << "<br>\n"; //length
      stream << tmp->text(3) << "<br>\n"; //info
      stream << "</td>\n";
      
      stream << "</tr>\n";
    }

    tmp = tmp->itemBelow();
  }

  stream << "</tbody>\n";
  stream << "</table>\n";
  stream << "</body>\n";
  stream << "</html>\n";

  file.close();

}


/*********************************************************
                 print playlist
*********************************************************/     

void PlayList::slotPrint()
{
  QImage logo;

  QString destPath = QDir::homeDirPath() + "/.kaffeine/kmp-playlist.png";
  
  if (!QFile::exists(destPath))
  {
    QStringList dirs = KGlobal::dirs()->findDirs("data","kaffeine");
    QString logoPath = dirs[0];  // global kde data path
    logoPath.append("kmp-playlist.png");
    logo.load(logoPath);
    logo.save(destPath, "PNG");
  }
  else
  {
    logo.load(destPath);
  }    

   //******get real name of user and date
  QString realName;
  QString date;

  QString uname = getenv("USER");
  if (!uname)
    uname = getenv("LOGNAME");
  if (uname)
  {
    struct passwd* pw = getpwnam(uname);
    realName = pw->pw_gecos;
  }

  date = KGlobal::locale()->formatDate(QDate::currentDate(), true);
  //**********
  

  QFont printFont = KGlobalSettings::generalFont();
  QFont boldPrintFont = printFont;
  boldPrintFont.setBold(true);
  QFontMetrics printFontMetrics(printFont);
  
  
  KPrinter* printer = new KPrinter;

  if (printer->setup(this))
  {
    printer->setFullPage(false);
    printer->setCreator("Kaffeine");
    printer->setDocName("Playlist");

    QPainter* p = new QPainter;
    p->begin(printer);
    p->setFont(printFont);
    p->setTabStops(6 * printFontMetrics.width("M"));

    QPen pen;
    pen.setWidth(2);
    p->setPen(pen);
    
    QPaintDeviceMetrics met(printer);

    p->drawImage(QPoint( (met.width()-logo.width())/2, 0), logo);
           

    int y = logo.height();
    int dy;
    QString text;
    
    text = realName + ", " + date;
    dy = p->boundingRect((met.width()-printFontMetrics.width(text))/2, y,
                         met.width(), met.height(), QPainter::ExpandTabs | QPainter::WordBreak,
                         text).height();
   
    p->drawText((met.width()-printFontMetrics.width(text))/2, y, met.width(),
                met.height()-y, QPainter::ExpandTabs | QPainter::WordBreak, text);
    y = y + dy + 20;

    QListViewItem* tmp = list->firstChild();

    while (tmp)
    {
      if (!tmp->parent())   //dont print sub urls
      {
        //******** title **********
        text = tmp->text(1);
        dy = p->boundingRect(0, y, met.width(), met.height(), QPainter::ExpandTabs | QPainter::WordBreak,
                            text).height();
        if(y+dy > met.height())
        {
          printer->newPage();
          y = 0;
        }
        p->setPen(QColor(0xFF8900));
        p->setFont(boldPrintFont);
        p->drawText(0, y, met.width(), met.height()-y, QPainter::ExpandTabs | QPainter::WordBreak,
                    text);
        p->setFont(printFont);            
        p->setPen(QColor("black"));
        y += dy;

        //******** url **********
        text = i18n("URL") + ":\t" + tmp->text(4);
        dy = p->boundingRect(0, y, met.width(), met.height(), QPainter::ExpandTabs | QPainter::WordBreak,
                            text).height();
        if(y+dy > met.height())
        {
          printer->newPage();
          y = 0;
        }
        p->drawText(0, y, met.width(), met.height()-y, QPainter::ExpandTabs | QPainter::WordBreak,
                    text);
        y += dy;
                        
        //******** length **********
        text = i18n("Length") + ":\t" + tmp->text(2);
        dy = p->boundingRect(0, y, met.width(), met.height(), QPainter::ExpandTabs | QPainter::WordBreak,
                             text).height();
        if(y+dy > met.height())
        {
          printer->newPage();
          y = 0;
        }
        p->drawText(0, y, met.width(), met.height()-y, QPainter::ExpandTabs | QPainter::WordBreak,
                    text);
        y += dy;

        //******** info **********
        text = i18n("Stream") + ":\t" + tmp->text(3);
        dy = p->boundingRect(0, y, met.width(), met.height(), QPainter::ExpandTabs | QPainter::WordBreak,
                             text).height();
        if(y+dy > met.height())
        {
          printer->newPage();
          y = 0;
        }
        p->drawText(0, y, met.width(), met.height()-y, QPainter::ExpandTabs | QPainter::WordBreak,
                    text);
        y += dy;

        //********** separator line **********
        dy = 10;
        if(y+dy > met.height())
        {
          printer->newPage();
          y = 0;
        }
        p->drawLine(0, y+4, met.width()-500, y+4);
        y += dy;
       
      }

     tmp = tmp->itemBelow();
    }

    p->end();
    delete p;
    
  }

  delete printer;
  
}

//********* something else....

static QString GetTimeString(int msec)
{
  int hours;
  int min;
  int sec;
  QString tmp;
  QString t;

    msec = msec/1000;  //sec
    hours = msec/3600;
    t = t.setNum(hours);
    t.append(":");

    msec = msec - (hours*3600);
    min = msec / 60;
    tmp = tmp.setNum(min);
    tmp = tmp.rightJustify(2, '0');
    t.append(tmp);
    t.append(":");

    sec = msec - (min*60);
    tmp = tmp.setNum(sec);
    tmp = tmp.rightJustify(2, '0');
    t.append(tmp);

   return t;
}


  
    

